from .constants import *
from .schema import Schema

class Pipe:
    def __init__(self, reg, yml):
        self.registry = reg
        self.source = yml
        self.assembly = []
        self.next_index = 0
        self.vars = {K_VAR: PKG_ID}
        self.data = {}

    def substitute(self, action):
        for key, value in action.items():
            if isinstance(value, str) and value[0] == K_VAR:
                var = value[1:]
                action[key] = self.vars[var]
        return action

    def init(self, action):
        self.registry.load(action['imports'])
        if action['set']:
            for k,v in action['set'].items(): self.vars[k] = v
        return self

    def compile(self):
        for id, action in self.source.items():
            action['id'] = id
            if action['do'] == K_INIT:
                self.init(action)
            else:
                self.substitute(action)
                asm = self.registry.assemble(action)
                self.assembly.append(asm)
            self.vars[K_VAR] = id
        return self.assembly

    def run(self):
        if len(self.assembly) == 0: self.compile()
        for da in self.assembly:
            method = da.CODE
            frame = method(self.data, da)
            self.data.setdefault(da.id, frame)
        return self.data



from .constants import *
from .load import *
from .schema import Schema

class Registry:
    def __init__(self, folder=None):
        self.schemas = {}
        self.modules = {}
        if not folder: folder = path_resource(PKG_ID, DAD_RESOURCE)
        self.add_folder(folder)

    def add_dad(self, name, yml):
        for act, dad in yml.items():
            key = f'{name}.{act}'
            self.schemas[key] = Schema(act, dad)

    def add_folder(self, folder):
        ydict = load_yamls(folder)
        for id, yml in ydict.items():
            name = id.replace(f'{DAD_RESOURCE}-','')
            self.add_dad(name, yml)

    def assemble(self, action):
        act = action['do']
        schema = self.schemas[act]
        action[K_CODE] = self.find_code(act)
        obj = schema.parse(action)
        return obj

    def load(self, imports):
        for key, value in imports.items():
            self.modules[key] = load_module(value)
        return self.modules

    def find_code(self, act):
        key, fn = act.split('.')
        mod = self.modules[key]
        if(hasattr(mod, fn)):
            func = getattr(mod, fn)
            return func
        return False



from collections import namedtuple
import json
from .constants import *

class Schema:
    def __init__(self, act, dad):
        self.act = act
        self.optional = ['doc','save',K_CODE]
        self.keys = ['id','do'] + self.optional
        self.dad = self.sanitize(dad)
        self.struct = namedtuple(self.act, self.keys)
        self.struct.__doc__ = json.dumps(dad)

    def sanitize(self, dad):
        newdad = {}
        for k, v in dad.items():
            key = k if k != K_FROM else K_FKEY
            newdad[key] = v
            self.keys.append(key)
            if v[-1] != '*': self.optional.append(key)
        return newdad

    def default(self, action):
        for k in self.optional:
            if k not in action: action[k] = False
        return action

    def parse(self, action):
        if K_FROM in action: action[K_FKEY] = action.pop(K_FROM)
        defaulted = self.default(action)
        tuple = self.struct(**defaulted)
        return tuple



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fridaay']

package_data = \
{'': ['*'], 'fridaay': ['dad/*', 'pipes/*']}

install_requires = \
['PyYAML>=6.0,<7.0', 'click>=8.1.3,<9.0.0', 'dad-sql-pandas>=0.1.1,<0.2.0']

setup_kwargs = {
    'name': 'fridaay',
    'version': '0.2.3',
    'description': 'Format Representing Interdependent Data Actions As YAML',
    'long_description': '# FRIDAAY\n## Format Representing Interdependent Data Actions As YAML\n\nWho needs SQL, Python, JavaScript and CSV?\nGet it all done by FRIDAAY\n\n# Usage\nFRIDAAY uses `poetry` to manage both dependencies and the virtual environment:\n```\n$ poetry install # or \'$ poetry update\'\n$ poetry env use python3\n$ poetry run pytest\n$ poetry run ptw\n```\n\n# Overview\n\nFRIDAAY defines a new "atomic unit" of abstraction for the modern data stack called Data Actions.  \nEach Data Action defines a semantic mapping for creating a new "frame" from existing frames (or inline data).\nThis allows analysts and data scientists to declaratively specify their intent, empowering the underlying platform to efficiently satisfy those requirements. We call this production-ready alternative to traditional exploratory notebooks a [PipeBook](https://ihack.us/2022/06/30/pipebook-yml-reimagining-notebooks-as-resilient-data-pipelines/).\n\nRight now, business logic and data dependencies are trapped inside complex (and often incompatible) programming languages such as SQL, Python, and Scala, and APIs like Spark vs Pandas, TensorFlow vs MLFlow, etc. FRIDAAY replaces these with a simple yet extensible "programming format" based on YAML that enables:\n- fine-grained orchestration\n- full-fidelity no-code visual programming of data pipelines\n- platform and language independence\n- reusable specification of dashboards and data apps  \n- inline tests and alerting\n- uniform specification of external integrations\n- schema-aware autocompletion and templates\n- ad-hoc materialization and incrementalism\n- version-controlled user-facing semantic models and metric layers\n- deterministic transformations between versions and vendors\n- novel interaction paradigms beyond notebooks and REPLs\n- turning legacy code into structured data, which we can manage using all our data superpowers\n\n## Example\nAvailable with the package in `folder = path_resource(PKG_ID, PIPE_FOLDER)`\n```\nfridaay:\n  version: 0.1\n  do: core.init\n  imports:\n   sql: dad_sql_pandas\n  set: # global constants (COMMENT)\n    NAME: demo_pets\n    SAPIENT: Human\n\ntest_data:\n  doc: Sample data for test purposes\n  do: sql.load\n  columns: [\'Name\',\'Age\',\'Weight\', \'Type\', \'Timestamp\']\n  data:\n  - [\'Ernie\', 54, 170.5, \'Human Tech Nerd\', 2020-03-20]\n  - [\'Qhuinn\', 7, 36.3, \'English Cocker Spaniel\', 2022-06-27]\n  - [\'Frolic\', 2, 76.2, \'Chocolate Labrador\', 2022-06-27]\n\ndemo_pets:\n  do: sql.select\n  from: $$ # last frame\n  cols:\n    Name: .str Personal Name\n    Age: .int.year Age\n    Weight: .float.pound Current Weight\n  where_all:\n  - ["Name","!=",Ernie]\n  #- [\'Timestamp\',\'>\', 2022-01-01]\n  save: [table]\n```\n\n# Releases\n```\n$ poetry version patch\n$ poetry build && poetry publish\n$ poetry version prepatch\n```\n',
    'author': 'Ernest Prabhakar',
    'author_email': 'ernest.prabhakar@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/TheSwanFactory/fridaay',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import os,sys,yaml,importlib
from .constants import *

def load_yaml(name, folder=PIPE_FOLDER):
    filename = f'{name}.yml'
    path = os.path.join(folder, filename)
    with open(path, 'r') as file:
        yml = yaml.safe_load(file)
        return yml

def load_yamls(folder=PIPE_FOLDER):
    ydict = {}
    for root, dirs, files in os.walk(folder):
       print(f'load_yamls: {root}')
       for name in files:
          split = name.split('.')
          key = split[0]
          if 'yml' in split: ydict[key] = load_yaml(key, root)
    return ydict

def load_module(module_name):
    mod = importlib.import_module(module_name)
    return mod

def path_package(package_name):
    posix_path = importlib.resources.files(package_name)
    return posix_path

def path_resource(package_name, subpath):
    for s in importlib.resources.contents(package_name):
        if (subpath in s):
            path = os.path.join(path_package(package_name), s)
            return path
